#include <SDL.h>
#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>
#include <unistd.h>
#include <string.h>

#include "def.h"
#include "log.h"
#include "physics.h"
#include "opengl.h"
#include "ship.h"
#include "pilot.h"
#include "player.h"
#include "joystick.h"
#include "space.h"
#include "rng.h"
#include "pilot.h"

#define CONF_FILE "conf"

static int quit = 0;

static unsigned int time = 0;

// Prototypes.

void print_usage(char** argv);

// Update.
static void update_all(void);

// Usage.
void print_usage(char** argv) {
  LOG("USAGE: %s [-f] [-j n | -J s] [-hv]", argv[0]);
  LOG("Options are:");
  LOG("\t-f   - Fullscreen");
  LOG("\t-w n - Set width to (n)");
  LOG("\t-h n - Set height to (n)");
  LOG("\t-j n - Use joystick (n)");
  LOG("\t-J s - Use joystick whose name contains (s)");
  LOG("\t-h   - Display this message and exit.");
  LOG("\t-v   - Print the version and exit");
}

int main(int argc, char** argv) {
  // Default values..
  gl_screen.w = 800;
  gl_screen.h = 640;
  gl_screen.fullscreen = 0;
  // Joystick.
  int indjoystick = -1;
  char* namjoystick = NULL;

  // Use Lua to parse configuration file.
  lua_State* L = luaL_newstate();
  if(luaL_dofile(L, CONF_FILE) == 0) {
    // OpenGL.
    lua_getglobal(L, "width");
    if(lua_isnumber(L, -1))
      gl_screen.w = (int)lua_tonumber(L, -1);
    lua_getglobal(L, "height");
    if(lua_isnumber(L, -1))
      gl_screen.h = (int)lua_tonumber(L, -1);
    lua_getglobal(L, "fullscreen");
    if(lua_isnumber(L, -1))
      if((int)lua_tonumber(L, -1) == 1)
        gl_screen.fullscreen = 1;

    // Joystick.
    lua_getglobal(L, "joystick");
    if(lua_isnumber(L, -1))
      indjoystick = (int)lua_tonumber(L, -1);
    else if(lua_isstring(L, -1))
      namjoystick = strdup((char*)lua_tostring(L, -1));
  }
  lua_close(L);

  // Parse arguments.
  int c = 0;
  while((c = getopt(argc, argv, "fJ:j:hv")) != -1) {
    switch(c) {
      case 'f':
        gl_screen.fullscreen = 1;
        break;
      case 'j':
        indjoystick = atoi(optarg);
        break;
      case 'J':
        namjoystick = strdup(optarg);
        break;
      case 'v':
        LOG("Lephisto: version %d.%d.%d\n", VMAJOR, VMINOR, VREV);
      case 'h':
        print_usage(argv);
        exit(EXIT_SUCCESS);
    }
  }

  // Random numbers.
  rng_init();

  // default window params.
  if(gl_init()) {
    // Initializes video output.
    WARN("Error initializing video output, exiting...");
    exit(EXIT_FAILURE);
  }

  // Input.
  if(indjoystick >= 0 || namjoystick != NULL) {
    if(joystick_init())
      WARN("Error initializing joystick input");
    if(namjoystick != NULL) {
      joystick_use(joystick_get(namjoystick));
      free(namjoystick);
    }
    else if(indjoystick >= 0)
      joystick_use(indjoystick);
  }

  // Data loading.
  ships_load();

  // Testing.
  unsigned int player_id;
  player_id = pilot_create(get_ship("Ship"), "Player", NULL, NULL, PILOT_PLAYER);
  gl_bindCamera(&get_pilot(player_id)->solid->pos);
  space_init();

  pilot_create(get_ship("Miss. Test"), NULL, NULL, NULL, 0);

  time = SDL_GetTicks();

  // Main looops.
  SDL_Event event;
  while(!quit) {
    // Event loop.
    while(SDL_PollEvent(&event)) {
      if(event.type == SDL_QUIT) quit = 1; // Handle quit.
      handle_input(&event);
    }
    update_all();
  }
  space_exit();
  // Unload data.
  pilots_free();
  ships_free();
  // Exit subsystems.
  joystick_exit();
  gl_exit(); // Kills video output.
  exit(EXIT_SUCCESS);
}

// Update all the things.
// Pilots:
//  -- Think (ai).
//  -- Solid.
static void update_all(void) {
  FP dt = (FP)(SDL_GetTicks() - time) / 1000.0;
  time = SDL_GetTicks();
  glClear(GL_COLOR_BUFFER_BIT);
  
  space_render(dt);
  pilots_update(dt);

  SDL_GL_SwapBuffers();
}

